@@ -2,10 +2,8 @@
 Python code for calculating non-parametric morphological diagnostics
 of galaxy images.
 """
-# Author: Vicente Rodriguez-Gomez <v.rodriguez@irya.unam.mx>
+# Author: Vicente Rodriguez-Gomez <vrodgom.astro@gmail.com>
 # Licensed under a 3-Clause BSD License.
-from __future__ import absolute_import, division, print_function
-
 from pkg_resources import parse_version
 import warnings
 import time
@@ -21,55 +19,81 @@ from astropy.utils import lazyproperty
 from astropy.stats import sigma_clipped_stats
 from astropy.modeling import models, fitting
 from astropy.utils.exceptions import AstropyUserWarning
+from astropy.convolution import convolve
 import photutils
 
 __all__ = ['ConvolvedSersic2D', 'SourceMorphology', 'source_morphology',
            '__version__']
 
-__version__ = '0.3.6'
+__version__ = '0.4.1'
+
+# A list of all the quantities that are calculated by SourceMorphology:
+_quantity_names = [
+    'xc_centroid',
+    'yc_centroid',
+    'ellipticity_centroid',
+    'elongation_centroid',
+    'orientation_centroid',
+    'xc_asymmetry',
+    'yc_asymmetry',
+    'ellipticity_asymmetry',
+    'elongation_asymmetry',
+    'orientation_asymmetry',
+    'flux_circ',
+    'flux_ellip',
+    'rpetro_circ',
+    'rpetro_ellip',
+    'rmax_circ',
+    'rmax_ellip',
+    'rhalf_circ',
+    'rhalf_ellip',
+    'r20',
+    'r50',
+    'r80',
+    'gini',
+    'm20',
+    'gini_m20_bulge',
+    'gini_m20_merger',
+    'sn_per_pixel',
+    'concentration',
+    'asymmetry',
+    'smoothness',
+    'multimode',
+    'intensity',
+    'deviation',
+    'outer_asymmetry',
+    'shape_asymmetry',
+    'sersic_amplitude',
+    'sersic_rhalf',
+    'sersic_n',
+    'sersic_xc',
+    'sersic_yc',
+    'sersic_ellip',
+    'sersic_theta',
+    'sky_mean',
+    'sky_median',
+    'sky_sigma',
+    'xmin_stamp',
+    'ymin_stamp',
+    'xmax_stamp',
+    'ymax_stamp',
+    'nx_stamp',
+    'ny_stamp',
+]
 
 def _quantile(sorted_values, q):
     """
-    For a sorted (in increasing order) 1-d array, return the
-    value corresponding to the quantile ``q``.
-    """
-    assert ((q >= 0) & (q <= 1))
-    if q == 1:
-        return sorted_values[-1]
-    else:
-        return sorted_values[int(q*len(sorted_values))]
-
-def _mode(a, axis=None):
-    """
-    Takes a masked array as input and returns the "mode"
-    as defined in Bertin & Arnouts (1996):
-    mode = 2.5 * median - 1.5 * mean
-    """
-    return 2.5*np.ma.median(a, axis=axis) - 1.5*np.ma.mean(a, axis=axis)
-
-def _local_variance(image):
-    """
-    Calculate a map of the local variance around each pixel, based on
-    its 8 adjacent neighbors
+    For a sorted (in increasing order) 1-d array, return the value
+    corresponding to the quantile ``q``.
 
     Notes
     -----
-    ndi.generic_filter(image, np.std, ...) is too slow,
-    so we do a workaround using ndi.convolve.
+    The result is identical to np.percentile(..., interpolation='lower'),
+    but the currently defined function is infinitely faster for sorted arrays.
     """
-    # Pixel weights, excluding central pixel.
-    w = np.array([
-        [1, 1, 1],
-        [1, 0, 1],
-        [1, 1, 1]], dtype=np.float64)
-    w = w / np.sum(w)
-
-    # Use the fact that var(x) = <x^2> - <x>^2.
-    local_mean = ndi.convolve(image, w)
-    local_mean2 = ndi.convolve(image**2, w)
-    local_var = local_mean2 - local_mean**2
-
-    return local_var
+    if q < 0 or q > 1:
+        raise ValueError('Quantiles must be in the range [0, 1].')
+    return sorted_values[int(q*(len(sorted_values)-1))]
 
 def _aperture_area(ap, mask, **kwargs):
     """
@@ -87,21 +111,17 @@ def _aperture_mean_nomask(ap, image, **kwargs):
     This avoids problems when the aperture is larger than the
     region of interest.
     """
-    if parse_version(photutils.__version__) <= parse_version('0.6'):
-        return ap.do_photometry(image, **kwargs)[0][0] / ap.area()
-    else:
-        return ap.do_photometry(image, **kwargs)[0][0] / ap.area
+    return ap.do_photometry(image, **kwargs)[0][0] / ap.area
 
 def _fraction_of_total_function_circ(r, image, center, fraction, total_sum):
     """
     Helper function to calculate ``_radius_at_fraction_of_total_circ``.
     """
-    if (r < 0) | (fraction < 0) | (fraction > 1) | (total_sum <= 0):
-        raise AssertionError
-    elif r == 0:
+    assert (r >= 0) & (fraction >= 0) & (fraction <= 1) & (total_sum > 0)
+    if r == 0:
         cur_fraction = 0.0
     else:
-        ap = photutils.CircularAperture(center, r)
+        ap = photutils.aperture.CircularAperture(center, r)
         # Force flux sum to be positive:
         ap_sum = np.abs(ap.do_photometry(image, method='exact')[0][0])
         cur_fraction = ap_sum / total_sum
@@ -115,13 +135,12 @@ def _radius_at_fraction_of_total_circ(image, center, r_total, fraction):
     """
     flag = 0  # flag=1 indicates a problem
 
-    ap_total = photutils.CircularAperture(center, r_total)
+    ap_total = photutils.aperture.CircularAperture(center, r_total)
 
     total_sum = ap_total.do_photometry(image, method='exact')[0][0]
-    if total_sum == 0:
-        raise AssertionError
-    elif total_sum < 0:
-        warnings.warn('Total flux sum is negative.', AstropyUserWarning)
+    assert total_sum != 0
+    if total_sum < 0:
+        warnings.warn('[r_circ] Total flux sum is negative.', AstropyUserWarning)
         flag = 1
         total_sum = np.abs(total_sum)
 
@@ -130,15 +149,11 @@ def _radius_at_fraction_of_total_circ(image, center, r_total, fraction):
     r_grid = np.linspace(0.0, r_total, num=npoints)
     i = 0  # initial value
     while True:
-        if i >= npoints:
-            raise Exception('Root not found within range.')
+        assert i < npoints, 'Root not found within range.'
         r = r_grid[i]
         curval = _fraction_of_total_function_circ(
             r, image, center, fraction, total_sum)
-        if curval == 0:
-            warnings.warn('Found root by pure chance!', AstropyUserWarning)
-            return r, flag
-        elif curval < 0:
+        if curval <= 0:
             r_min = r
         elif curval > 0:
             r_max = r
@@ -155,13 +170,12 @@ def _fraction_of_total_function_ellip(a, image, center, elongation, theta,
     """
     Helper function to calculate ``_radius_at_fraction_of_total_ellip``.
     """
-    if (a < 0) | (fraction < 0) | (fraction > 1) | (total_sum <= 0):
-        raise AssertionError
-    elif a == 0:
+    assert (a >= 0) & (fraction >= 0) & (fraction <= 1) & (total_sum > 0)
+    if a == 0:
         cur_fraction = 0.0
     else:
         b = a / elongation
-        ap = photutils.EllipticalAperture(center, a, b, theta=theta)
+        ap = photutils.aperture.EllipticalAperture(center, a, b, theta=theta)
         # Force flux sum to be positive:
         ap_sum = np.abs(ap.do_photometry(image, method='exact')[0][0])
         cur_fraction = ap_sum / total_sum
@@ -178,14 +192,13 @@ def _radius_at_fraction_of_total_ellip(image, center, elongation, theta,
     flag = 0  # flag=1 indicates a problem
 
     b_total = a_total / elongation
-    ap_total = photutils.EllipticalAperture(
+    ap_total = photutils.aperture.EllipticalAperture(
         center, a_total, b_total, theta=theta)
 
     total_sum = ap_total.do_photometry(image, method='exact')[0][0]
-    if total_sum == 0:
-        raise AssertionError
-    elif total_sum < 0:
-        warnings.warn('Total flux sum is negative.', AstropyUserWarning)
+    assert total_sum != 0
+    if total_sum < 0:
+        warnings.warn('[r_ellip] Total flux sum is negative.', AstropyUserWarning)
         flag = 1
         total_sum = np.abs(total_sum)
 
@@ -194,15 +207,11 @@ def _radius_at_fraction_of_total_ellip(image, center, elongation, theta,
     a_grid = np.linspace(0.0, a_total, num=npoints)
     i = 0  # initial value
     while True:
-        if i >= npoints:
-            raise Exception('Root not found within range.')
+        assert i < npoints, 'Root not found within range.'
         a = a_grid[i]
         curval = _fraction_of_total_function_ellip(
             a, image, center, elongation, theta, fraction, total_sum)
-        if curval == 0:
-            warnings.warn('Found root by pure chance!', AstropyUserWarning)
-            return a, flag
-        elif curval < 0:
+        if curval <= 0:
             a_min = a
         elif curval > 0:
             a_max = a
@@ -242,13 +251,12 @@ class ConvolvedSersic2D(models.Sersic2D):
         z_sersic = models.Sersic2D.evaluate(x, y, amplitude, r_eff, n, x_0, y_0,
                                             ellip, theta)
         if cls.psf is None:
-            raise Exception('Must specify PSF using set_psf method.')
+            raise AssertionError('Must specify PSF using set_psf method.')
 
         # Apparently, scipy.signal also wants double:
         return scipy.signal.fftconvolve(
             np.float64(z_sersic), np.float64(cls.psf), mode='same')
 
-
 class SourceMorphology(object):
     """
     Class to measure the morphological parameters of a single labeled
@@ -259,7 +267,7 @@ class SourceMorphology(object):
     image : array-like
         A 2D image containing the sources of interest.
         The image must already be background-subtracted.
-    segmap : array-like (int) or ``photutils.SegmentationImage``
+    segmap : array-like (int) or ``photutils.segmentation.SegmentationImage``
         A 2D segmentation map where different sources are
         labeled with different positive integer values.
         A value of zero is reserved for the background.
@@ -353,6 +361,10 @@ class SourceMorphology(object):
         The minimum ratio between the area of the intersection of
         all 3 segmaps and the area of the largest segmap in order to
         have a good measurement.
+    verbose : bool, optional
+        If ``True``, this prints various minor warnings (which do not
+        set off the "bad measurement" flag) during the calculations.
+        The default value is ``False``.
 
     References
     ----------
@@ -366,7 +378,7 @@ class SourceMorphology(object):
                  petro_extent_cas=1.5, petro_fraction_cas=0.25,
                  boxcar_size_mid=3.0, niter_bh_mid=5, sigma_mid=1.0,
                  petro_extent_flux=2.0, boxcar_size_shape_asym=3.0,
-                 sersic_maxiter=500, segmap_overlap_ratio=0.25):
+                 sersic_maxiter=500, segmap_overlap_ratio=0.25, verbose=False):
         self._image = image
         self._segmap = segmap
         self.label = label
@@ -390,79 +402,26 @@ class SourceMorphology(object):
         self._boxcar_size_shape_asym = boxcar_size_shape_asym
         self._sersic_maxiter = sersic_maxiter
         self._segmap_overlap_ratio = segmap_overlap_ratio
-
-        # A list of all the properties that will be calculated:
-        self._quantity_names = [
-            'xc_centroid',
-            'yc_centroid',
-            'ellipticity_centroid',
-            'elongation_centroid',
-            'orientation_centroid',
-            'xc_asymmetry',
-            'yc_asymmetry',
-            'ellipticity_asymmetry',
-            'elongation_asymmetry',
-            'orientation_asymmetry',
-            'flux_circ',
-            'flux_ellip',
-            'rpetro_circ',
-            'rpetro_ellip',
-            'rmax_circ',
-            'rmax_ellip',
-            'rhalf_circ',
-            'rhalf_ellip',
-            'r20',
-            'r50',
-            'r80',
-            'gini',
-            'm20',
-            'gini_m20_bulge',
-            'gini_m20_merger',
-            'sn_per_pixel',
-            'concentration',
-            'asymmetry',
-            'smoothness',
-            'multimode',
-            'intensity',
-            'deviation',
-            'outer_asymmetry',
-            'shape_asymmetry',
-            'sersic_amplitude',
-            'sersic_rhalf',
-            'sersic_n',
-            'sersic_xc',
-            'sersic_yc',
-            'sersic_ellip',
-            'sersic_theta',
-            'sky_mean',
-            'sky_median',
-            'sky_sigma',
-            'xmin_stamp',
-            'ymin_stamp',
-            'xmax_stamp',
-            'ymax_stamp',
-            'nx_stamp',
-            'ny_stamp',
-        ]
+        self._verbose = verbose
 
         # Measure runtime
         start = time.time()
 
-        if not isinstance(self._segmap, photutils.SegmentationImage):
-            self._segmap = photutils.SegmentationImage(self._segmap)
+        if not isinstance(self._segmap, photutils.segmentation.SegmentationImage):
+            self._segmap = photutils.segmentation.SegmentationImage(self._segmap)
 
         # Check sanity of input data
-        if parse_version(photutils.__version__) < parse_version('0.5'):
-            self._segmap.check_label(self.label)
-        else:
-            self._segmap.check_labels([self.label])
+        self._segmap.check_labels([self.label])
         assert self._segmap.data.shape == self._image.shape
         if self._mask is not None:
             assert self._mask.shape == self._image.shape
-            assert self._mask.dtype == np.bool8
+            assert self._mask.dtype == np.bool_
         if self._weightmap is not None:
             assert self._weightmap.shape == self._image.shape
 
+        if self._weightmap is None and self._gain is None:
+            raise AssertionError('Must provide either weightmap or gain.')
+
         # Normalize PSF
         if self._psf is not None:
             self._psf = self._psf / np.sum(self._psf)
@@ -490,7 +449,7 @@ class SourceMorphology(object):
         self._yc_stamp = self.yc_centroid - self.ymin_stamp
 
         # Print warning if centroid is masked:
-        ic, jc = int(self._yc_stamp), int(self._xc_stamp)
+        ic, jc = int(np.round(self._yc_stamp)), int(np.round(self._xc_stamp))
         if self._cutout_stamp_maskzeroed[ic, jc] == 0:
             warnings.warn('Centroid is masked.', AstropyUserWarning)
             self.flag = 1
@@ -540,8 +499,10 @@ class SourceMorphology(object):
         w = w / np.sum(w)
 
         # Use the fact that var(x) = <x^2> - <x>^2.
-        local_mean = ndi.convolve(image, w)
-        local_mean2 = ndi.convolve(image**2, w)
+        local_mean = convolve(image, w, boundary='extend',
+                              normalize_kernel=False)
+        local_mean2 = convolve(image**2, w, boundary='extend',
+                               normalize_kernel=False)
         local_std = np.sqrt(local_mean2 - local_mean**2)
 
         # Get "bad pixels"
@@ -557,7 +518,7 @@ class SourceMorphology(object):
         deal with these cases by creating an "empty" object and
         interrupting the constructor.
         """
-        for q in self._quantity_names:
+        for q in _quantity_names:
             setattr(self, q, -99.0)
         self.nx_stamp = -99
         self.ny_stamp = -99
@@ -571,7 +532,7 @@ class SourceMorphology(object):
         Calculate all morphological parameters, which are stored
         as "lazy" properties.
         """
-        for q in self._quantity_names:
+        for q in _quantity_names:
             tmp = self[q]
 
     def _check_segmaps(self):
@@ -849,7 +810,7 @@ class SourceMorphology(object):
         Flag badpixels (outliers).
         """
         self.num_badpixels = -1
-        badpixels = np.zeros((self.ny_stamp, self.nx_stamp), dtype=np.bool8)
+        badpixels = np.zeros((self.ny_stamp, self.nx_stamp), dtype=np.bool_)
         if self._n_sigma_outlier > 0:
             badpixels = self._get_badpixels(self._image[self._slice_stamp])
             self.num_badpixels = np.sum(badpixels)
@@ -910,12 +871,10 @@ class SourceMorphology(object):
         ``gain`` argument.
         """
         if self._weightmap is None:
-            if self._gain is None:
-                raise Exception('Must provide either weightmap or gain.')
-            else:
-                assert self._gain > 0
-                weightmap_stamp = np.sqrt(
-                    np.abs(self._image[self._slice_stamp])/self._gain + self.sky_sigma**2)
+            # Already checked that gain is not None:
+            assert self._gain > 0
+            weightmap_stamp = np.sqrt(
+                np.abs(self._image[self._slice_stamp])/self._gain + self.sky_sigma**2)
         else:
             weightmap_stamp = self._weightmap[self._slice_stamp]
 
@@ -971,8 +930,8 @@ class SourceMorphology(object):
         r_in = r - 0.5 * self._annulus_width
         r_out = r + 0.5 * self._annulus_width
 
-        circ_annulus = photutils.CircularAnnulus(center, r_in, r_out)
-        circ_aperture = photutils.CircularAperture(center, r)
+        circ_annulus = photutils.aperture.CircularAnnulus(center, r_in, r_out)
+        circ_aperture = photutils.aperture.CircularAperture(center, r)
 
         # Force mean fluxes to be positive:
         circ_annulus_mean_flux = np.abs(_aperture_mean_nomask(
@@ -982,7 +941,9 @@ class SourceMorphology(object):
 
         if circ_aperture_mean_flux == 0:
             warnings.warn('[rpetro_circ] Mean flux is zero.', AstropyUserWarning)
-            ratio = 1.0
+            # If flux within annulus is also zero (e.g. beyond the image
+            # boundaries), return zero. Otherwise return 1.0:
+            ratio = float(circ_annulus_mean_flux != 0)
             self.flag = 1
         else:
             ratio = circ_annulus_mean_flux / circ_aperture_mean_flux
@@ -1009,30 +970,31 @@ class SourceMorphology(object):
         r_inner = self._annulus_width
         r_outer = self._diagonal_distance
         assert r_inner < r_outer
-        dr = (r_outer - r_inner) / float(npoints-1)
         r_min, r_max = None, None
-        r = r_inner  # initial value
-        while True:
-            if r >= r_outer:
-                warnings.warn('[rpetro_circ] rpetro larger than cutout.',
-                              AstropyUserWarning)
-                self.flag = 1
+        for r in np.linspace(r_inner, r_outer, npoints):
             curval = self._petrosian_function_circ(r, center)
             if curval == 0:
-                warnings.warn('[rpetro_circ] We found rpetro by pure chance!',
+                warnings.warn('[rpetro_circ] Found rpetro by chance?',
                               AstropyUserWarning)
                 return r
-            elif curval > 0:
+            elif curval > 0:  # we have not reached rpetro yet
                 r_min = r
-            elif curval < 0:
+            else:  # we are beyond rpetro
                 if r_min is None:
-                    warnings.warn('[rpetro_circ] r_min is not defined yet.',
+                    warnings.warn('rpetro_circ < annulus_width! ' +
+                                  'Setting rpetro_circ = annulus_width.',
                                   AstropyUserWarning)
                     self.flag = 1
-                else:
-                    r_max = r
-                    break
-            r += dr
+                    return r_inner
+                r_max = r
+                break
+
+        assert r_min is not None
+        if r_max is None:
+            warnings.warn('[rpetro_circ] rpetro too large.',
+                          AstropyUserWarning)
+            self.flag = 1
+            r_max = r_outer
 
         rpetro_circ = opt.brentq(self._petrosian_function_circ,
                                  r_min, r_max, args=(center,), xtol=1e-6)
@@ -1066,7 +1028,7 @@ class SourceMorphology(object):
         """
         image = self._cutout_stamp_maskzeroed
         r = self._petro_extent_flux * self.rpetro_circ
-        ap = photutils.CircularAperture(self._asymmetry_center, r)
+        ap = photutils.aperture.CircularAperture(self._asymmetry_center, r)
         # Force flux sum to be positive:
         ap_sum = np.abs(ap.do_photometry(image, method='exact')[0][0])
         return ap_sum
@@ -1090,9 +1052,9 @@ class SourceMorphology(object):
 
         b_out = a_out / elongation
 
-        ellip_annulus = photutils.EllipticalAnnulus(
+        ellip_annulus = photutils.aperture.EllipticalAnnulus(
             center, a_in, a_out, b_out, theta=theta)
-        ellip_aperture = photutils.EllipticalAperture(
+        ellip_aperture = photutils.aperture.EllipticalAperture(
             center, a, b, theta=theta)
 
         # Force mean fluxes to be positive:
@@ -1103,7 +1065,9 @@ class SourceMorphology(object):
 
         if ellip_aperture_mean_flux == 0:
             warnings.warn('[rpetro_ellip] Mean flux is zero.', AstropyUserWarning)
-            ratio = 1.0
+            # If flux within annulus is also zero (e.g. beyond the image
+            # boundaries), return zero. Otherwise return 1.0:
+            ratio = float(ellip_annulus_mean_flux != 0)
             self.flag = 1
         else:
             ratio = ellip_annulus_mean_flux / ellip_aperture_mean_flux
@@ -1129,32 +1093,33 @@ class SourceMorphology(object):
         # Find appropriate range for root finder
         npoints = 100
         a_inner = self._annulus_width
-        a_outer = self._diagonal_distance
+        a_outer = self._diagonal_distance * elongation
         assert a_inner < a_outer
-        da = (a_outer - a_inner) / float(npoints-1)
         a_min, a_max = None, None
-        a = a_inner  # initial value
-        while True:
-            if a >= a_outer:
-                warnings.warn('[rpetro_ellip] rpetro larger than cutout.',
-                              AstropyUserWarning)
-                self.flag = 1
+        for a in np.linspace(a_inner, a_outer, npoints):
             curval = self._petrosian_function_ellip(a, center, elongation, theta)
             if curval == 0:
-                warnings.warn('[rpetro_ellip] We found rpetro by pure chance!',
+                warnings.warn('[rpetro_ellip] Found rpetro by chance?',
                               AstropyUserWarning)
                 return a
-            elif curval > 0:
+            elif curval > 0:  # we have not reached rpetro yet
                 a_min = a
-            elif curval < 0:
+            else:  # we are beyond rpetro
                 if a_min is None:
-                    warnings.warn('[rpetro_ellip] a_min is not defined yet.',
+                    warnings.warn('rpetro_ellip < annulus_width! ' +
+                                  'Setting rpetro_ellip = annulus_width.',
                                   AstropyUserWarning)
                     self.flag = 1
-                else:
-                    a_max = a
-                    break
-            a += da
+                    return a_inner
+                a_max = a
+                break
+
+        assert a_min is not None
+        if a_max is None:
+            warnings.warn('[rpetro_ellip] rpetro too large.',
+                          AstropyUserWarning)
+            self.flag = 1
+            a_max = a_outer
 
         rpetro_ellip = opt.brentq(self._petrosian_function_ellip, a_min, a_max,
                                   args=(center, elongation, theta,), xtol=1e-6)
@@ -1182,7 +1147,7 @@ class SourceMorphology(object):
         a = self._petro_extent_flux * self.rpetro_ellip
         b = a / self.elongation_asymmetry
         theta = self.orientation_asymmetry
-        ap = photutils.EllipticalAperture(
+        ap = photutils.aperture.EllipticalAperture(
             self._asymmetry_center, a, b, theta=theta)
         # Force flux sum to be positive:
         ap_sum = np.abs(ap.do_photometry(image, method='exact')[0][0])
@@ -1207,7 +1172,7 @@ class SourceMorphology(object):
         a_out = self.rpetro_ellip + 0.5 * self._annulus_width
         b_out = a_out / self.elongation_asymmetry
         theta = self.orientation_asymmetry
-        ellip_annulus = photutils.EllipticalAnnulus(
+        ellip_annulus = photutils.aperture.EllipticalAnnulus(
             (self._xc_stamp, self._yc_stamp), a_in, a_out, b_out, theta=theta)
         ellip_annulus_mean_flux = _aperture_mean_nomask(
             ellip_annulus, cutout_smooth, method='exact')
@@ -1346,6 +1311,29 @@ class SourceMorphology(object):
 
         return 0.139*self.m20 + 0.990*self.gini - 0.327
 
+    # @lazyproperty
+    # def sn_per_pixel(self):
+    #     """
+    #     Calculate the signal-to-noise per pixel using the Petrosian segmap.
+    #     """
+    #     weightmap = self._weightmap_stamp
+    #     if np.any(weightmap < 0):
+    #         warnings.warn('[sn_per_pixel] Some negative weightmap values.',
+    #                       AstropyUserWarning)
+    #         weightmap = np.abs(weightmap)
+    #
+    #     locs = self._segmap_gini & (self._cutout_stamp_maskzeroed >= 0)
+    #     pixelvals = self._cutout_stamp_maskzeroed[locs]
+    #     # The sky background noise is already included in the weightmap:
+    #     snp = np.mean(pixelvals / weightmap[locs])
+    #
+    #     if not np.isfinite(snp):
+    #         warnings.warn('Invalid sn_per_pixel.', AstropyUserWarning)
+    #         self.flag = 1
+    #         snp = -99.0  # invalid
+    #
+    #     return snp
+
     @lazyproperty
     def sn_per_pixel(self):
         """
@@ -1357,15 +1345,16 @@ class SourceMorphology(object):
                           AstropyUserWarning)
             weightmap = np.abs(weightmap)
 
-        locs = self._segmap_gini & (self._cutout_stamp_maskzeroed >= 0)
-        pixelvals = self._cutout_stamp_maskzeroed[locs]
-        # The sky background noise is already included in the weightmap:
-        snp = np.mean(pixelvals / weightmap[locs])
-
-        if not np.isfinite(snp):
+        locs = (self._segmap_gini & (self._cutout_stamp_maskzeroed >= 0) &
+                (weightmap > 0))
+        if np.sum(locs) == 0:
             warnings.warn('Invalid sn_per_pixel.', AstropyUserWarning)
             self.flag = 1
             snp = -99.0  # invalid
+        else:
+            pixelvals = self._cutout_stamp_maskzeroed[locs]
+            # The sky background noise is already included in the weightmap:
+            snp = np.mean(pixelvals / weightmap[locs])
 
         return snp
 
@@ -1383,10 +1372,11 @@ class SourceMorphology(object):
         """
         segmap = self._segmap.data[self._slice_stamp]
         ny, nx = segmap.shape
-        mask = np.zeros(segmap.shape, dtype=np.bool8)
+        mask = np.zeros(segmap.shape, dtype=np.bool_)
         if self._mask is not None:
             mask = self._mask[self._slice_stamp]
 
+        assert self._skybox_size >= 2
         cur_skybox_size = self._skybox_size
         while True:
             for i in range(0, ny - cur_skybox_size):
@@ -1397,17 +1387,15 @@ class SourceMorphology(object):
                         return boxslice
 
             # If we got here, a skybox of the given size was not found.
-            if cur_skybox_size <= 1:
+            if cur_skybox_size <= 2:
                 warnings.warn('[skybox] Skybox not found.', AstropyUserWarning)
                 self.flag = 1
                 return slice(0, 0), slice(0, 0)
             else:
                 cur_skybox_size //= 2
-                warnings.warn('[skybox] Reducing skybox size to %d.' % (
-                              cur_skybox_size), AstropyUserWarning)
-
-        # Should not reach this point.
-        raise AssertionError
+                if self._verbose:
+                    warnings.warn('[skybox] Reducing skybox size to %d.' % (
+                                  cur_skybox_size), AstropyUserWarning)
 
     @lazyproperty
     def sky_mean(self):
@@ -1527,27 +1515,23 @@ class SourceMorphology(object):
         # Create aperture for the chosen kind of asymmetry
         if kind == 'cas':
             r = self._petro_extent_cas * self._rpetro_circ_centroid
-            ap = photutils.CircularAperture(center, r)
+            ap = photutils.aperture.CircularAperture(center, r)
         elif kind == 'outer':
             a_in = self.rhalf_ellip
             a_out = self.rmax_ellip
             b_out = a_out / self.elongation_asymmetry
             theta = self.orientation_asymmetry
-            if np.isnan(a_in) or np.isnan(a_out) or (a_in <= 0) or (a_out <= 0):
-                warnings.warn('[outer_asym] Invalid annulus dimensions.',
-                              AstropyUserWarning)
-                self.flag = 1
-                return -99.0  # invalid
-            ap = photutils.EllipticalAnnulus(center, a_in, a_out, b_out, theta=theta)
+            assert (a_in > 0) & (a_out > 0)
+            ap = photutils.aperture.EllipticalAnnulus(center, a_in, a_out, b_out, theta=theta)
         elif kind == 'shape':
             if np.isnan(self.rmax_circ) or (self.rmax_circ <= 0):
                 warnings.warn('[shape_asym] Invalid rmax_circ value.',
                               AstropyUserWarning)
                 self.flag = 1
                 return -99.0  # invalid
-            ap = photutils.CircularAperture(center, self.rmax_circ)
+            ap = photutils.aperture.CircularAperture(center, self.rmax_circ)
         else:
-            raise Exception('Asymmetry kind not understood:', kind)
+            raise NotImplementedError('Asymmetry kind not understood:', kind)
 
         # Apply eq. 10 from Lotz et al. (2004)
         ap_abs_sum = ap.do_photometry(np.abs(image), method='exact')[0][0]
@@ -1589,7 +1573,7 @@ class SourceMorphology(object):
             center_asym = center_0
 
         # Print warning if center is masked
-        ic, jc = int(center_asym[1]), int(center_asym[0])
+        ic, jc = int(np.round(center_asym[1])), int(np.round(center_asym[0]))
         if self._cutout_stamp_maskzeroed[ic, jc] == 0:
             warnings.warn('[asym_center] Asymmetry center is masked.',
                           AstropyUserWarning)
@@ -1727,15 +1711,16 @@ class SourceMorphology(object):
     @lazyproperty
     def smoothness(self):
         """
-        Calculate smoothness (a.k.a. clumpiness) as described in
-        Conselice (2003).
+        Calculate smoothness (a.k.a. clumpiness) as defined in eq. (11)
+        from Lotz et al. (2004). Note that the original definition by
+        Conselice (2003) includes an additional factor of 10.
         """
         image = self._cutout_stamp_maskzeroed
 
         # Exclude central region during smoothness calculation:
         r_in = self._petro_fraction_cas * self.rpetro_circ
         r_out = self._petro_extent_cas * self.rpetro_circ
-        ap = photutils.CircularAnnulus(self._asymmetry_center, r_in, r_out)
+        ap = photutils.aperture.CircularAnnulus(self._asymmetry_center, r_in, r_out)
 
         boxcar_size = int(self._petro_fraction_cas * self.rpetro_circ)
         image_smooth = ndi.uniform_filter(image, size=boxcar_size)
@@ -1755,10 +1740,7 @@ class SourceMorphology(object):
         if self._sky_smoothness == -99.0:  # invalid skybox
             S = ap_diff / ap_flux
         else:
-            if parse_version(photutils.__version__) <= parse_version('0.6'):
-                S = (ap_diff - ap.area()*self._sky_smoothness) / ap_flux
-            else:
-                S = (ap_diff - ap.area*self._sky_smoothness) / ap_flux
+            S = (ap_diff - ap.area*self._sky_smoothness) / ap_flux
 
         if not np.isfinite(S):
             warnings.warn('Invalid smoothness.', AstropyUserWarning)
@@ -1782,8 +1764,8 @@ class SourceMorphology(object):
 
         # Instead of assuming that the main segment is at the center
         # of the stamp, use the position of the brightest pixel:
-        ic = int(self._y_maxval_stamp)
-        jc = int(self._x_maxval_stamp)
+        ic = int(np.round(self._y_maxval_stamp))
+        jc = int(np.round(self._x_maxval_stamp))
 
         # Grow regions using 8-connected neighbor "footprint"
         s = ndi.generate_binary_structure(2, 2)
@@ -1840,8 +1822,10 @@ class SourceMorphology(object):
         # In these cases we simply assume that the MID segmap
         # is equal to the Gini segmap.
         if self._segmap_mid_function(0.0) > 0.0:
-            warnings.warn('segmap_mid is undefined; using segmap_gini instead.',
-                          AstropyUserWarning)
+            if self._verbose:
+                warnings.warn(
+                    'segmap_mid is undefined; using segmap_gini instead.',
+                    AstropyUserWarning)
             return self._segmap_gini
 
         # Find appropriate quantile using numerical solver
@@ -1858,8 +1842,8 @@ class SourceMorphology(object):
         segmap = segmap_float > 0.5
 
         # Make sure that brightest pixel is in segmap
-        ic = int(self._y_maxval_stamp)
-        jc = int(self._x_maxval_stamp)
+        ic = int(np.round(self._y_maxval_stamp))
+        jc = int(np.round(self._x_maxval_stamp))
         if ~segmap[ic, jc]:
             warnings.warn('[segmap_mid] Adding brightest pixel to segmap.',
                           AstropyUserWarning)
@@ -1983,14 +1967,17 @@ class SourceMorphology(object):
             if ratio_min < 0:  # valid "ratios" should be negative
                 break
             elif mid_stepsize < 1e-3:
-                warnings.warn('[M statistic] Single clump!', AstropyUserWarning)
+                if self._verbose:
+                    warnings.warn(
+                        '[M statistic] Single clump!', AstropyUserWarning)
                 # This sometimes happens when the source is a star, so the user
                 # might want to discard M=0 cases, depending on the dataset.
                 return 0.0
             else:
                 mid_stepsize = mid_stepsize / 2.0
-                warnings.warn('[M statistic] Reduced stepsize to %g.' % (
-                              mid_stepsize), AstropyUserWarning)
+                if self._verbose:
+                    warnings.warn('[M statistic] Reduced stepsize to %g.' % (
+                                  mid_stepsize), AstropyUserWarning)
 
         # STAGE 2: basin-hopping method
 
@@ -2030,7 +2017,7 @@ class SourceMorphology(object):
         Returns a labeled array indicating regions around local maxima.
         """
         peaks = skimage.feature.peak_local_max(
-            self._cutout_mid_smooth, indices=True, num_peaks=np.inf)
+            self._cutout_mid_smooth, num_peaks=np.inf)
         num_peaks = peaks.shape[0]
         # The zero label is reserved for the background:
         peak_labels = np.arange(1, num_peaks+1, dtype=np.int64)
@@ -2189,13 +2176,10 @@ class SourceMorphology(object):
         # that only contains background sky (hopefully).
         r_in = self._petro_extent_flux * self.rpetro_ellip
         r_out = 2.0 * self._petro_extent_flux * self.rpetro_ellip
-        circ_annulus = photutils.CircularAnnulus(center, r_in, r_out)
+        circ_annulus = photutils.aperture.CircularAnnulus(center, r_in, r_out)
 
         # Convert circular annulus aperture to binary mask
-        if parse_version(photutils.__version__) <= parse_version('0.6'):
-            circ_annulus_mask = circ_annulus.to_mask(method='center')[0]
-        else:
-            circ_annulus_mask = circ_annulus.to_mask(method='center')
+        circ_annulus_mask = circ_annulus.to_mask(method='center')
 
         # With the same shape as the postage stamp
         circ_annulus_mask = circ_annulus_mask.to_image((ny, nx))
@@ -2205,9 +2189,10 @@ class SourceMorphology(object):
         # If sky area is too small (e.g., if annulus is outside the
         # image), use skybox instead.
         if np.sum(~total_mask) < self._skybox_size**2:
-            warnings.warn('[shape_asym] Using skybox for background.',
-                          AstropyUserWarning)
-            total_mask = np.ones((ny, nx), dtype=np.bool8)
+            if self._verbose:
+                warnings.warn('[shape_asym] Using skybox for background.',
+                              AstropyUserWarning)
+            total_mask = np.ones((ny, nx), dtype=np.bool_)
             total_mask[self._slice_skybox] = False
             # However, if skybox is undefined, there is nothing to do.
             if np.sum(~total_mask) == 0:
@@ -2216,10 +2201,14 @@ class SourceMorphology(object):
                 self.flag = 1
                 return ~self._mask_stamp_no_bg
 
+        # Define the "mode" as in Bertin & Arnouts (1996):
+        bkg_estimator = photutils.background.ModeEstimatorBackground(
+            median_factor=2.5, mean_factor=1.5)
+
         # Do sigma-clipping until convergence
         mean, median, std = sigma_clipped_stats(
             self._cutout_stamp_maskzeroed, mask=total_mask, sigma=3.0,
-            maxiters=None, cenfunc=_mode)
+            maxiters=None, cenfunc=bkg_estimator)
 
         # Mode as defined in Bertin & Arnouts (1996)
         mode = 2.5*median - 1.5*mean
@@ -2346,7 +2335,7 @@ class SourceMorphology(object):
             self.flag_sersic = 1
             a_in = self.rhalf_ellip
         b_out = a_out / self.elongation_asymmetry
-        ellip_annulus = photutils.EllipticalAnnulus(
+        ellip_annulus = photutils.aperture.EllipticalAnnulus(
             center, a_in, a_out, b_out, theta=theta)
         ellip_annulus_mean_flux = _aperture_mean_nomask(
             ellip_annulus, image, method='exact')
@@ -2366,12 +2355,8 @@ class SourceMorphology(object):
         fit_weights[locs] = 1.0 / weightmap[locs]
 
         # Initial guess
-        if self.concentration < 3.0:
-            guess_n = 1.0
-        elif self.concentration < 4.0:
-            guess_n = 2.0
-        else:
-            guess_n = 3.5
+        guess_n = 10.0**(-1.5) * self.concentration**3.5  # empirical
+        guess_n = min(max(guess_n, 1.0), 3.5)  # limit to range [1.0, 3.5]
         xc, yc = self._asymmetry_center
         if self._psf is None:
             sersic_init = models.Sersic2D(
@@ -2479,7 +2464,7 @@ def source_morphology(image, segmap, **kwargs):
     image : array-like
         A 2D image containing the sources of interest.
         The image must already be background-subtracted.
-    segmap : array-like (int) or `photutils.SegmentationImage`
+    segmap : array-like (int) or `photutils.segmentation.SegmentationImage`
         A 2D segmentation map where different sources are
         labeled with different positive integer values.
         A value of zero is reserved for the background.
@@ -2497,7 +2482,7 @@ def source_morphology(image, segmap, **kwargs):
 
     See Also
     --------
-    `SourceMorphology` : Class to measure morphological parameters.
+    SourceMorphology : Class to measure morphological parameters.
 
     Examples
     --------
@@ -2508,12 +2493,11 @@ def source_morphology(image, segmap, **kwargs):
     See `README.rst` for a list of references.
 
     """
-    if not isinstance(segmap, photutils.SegmentationImage):
-        segmap = photutils.SegmentationImage(segmap)
+    if not isinstance(segmap, photutils.segmentation.SegmentationImage):
+        segmap = photutils.segmentation.SegmentationImage(segmap)
 
     sources_morph = []
     for label in segmap.labels:
         sources_morph.append(SourceMorphology(image, segmap, label, **kwargs))
-        print('Finished processing source %d.\n' % (label,))
 
     return sources_morph
